/* Copyright IBM Corp. 2017 */
import { join } from 'path';
import { combineLatest, empty, from, merge, Observable, of, Subject } from 'rxjs';
import { map, mergeMap, tap, toArray } from 'rxjs/operators';
import { parse as urlParse } from 'url';

import { firstOf } from '../utils/file.utils';
import {
    CMD_ADD,
    OPTION_DATA,
    OPTION_NAME,
    OPTION_RELATIVE,
    OPTION_ROOT,
    OPTION_SCRIPT,
    OPTION_SOURCE,
    OPTION_VERBOSE,
} from './../constants';
import { Options, optNormalizeDir } from './../options';
import { configIndexDirFromContextRoot, configTrimSlashes } from './../utils/config';
import { hashString, hashToIdentifier, longHash } from './../utils/hash.utils';
import { hbsCompile } from './../utils/rx.hbs';
import { readFile, writeFile } from './../utils/rx.walk';
import { SiteNode } from './../utils/wchtools';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';
import {
    HELP_APPLICATION_NAME,
    HELP_CONTEXT_ROOT,
    HELP_COPY_SOURCE,
    HELP_DATA_DIRECTORY,
    HELP_RELATIVE,
    HELP_VERBOSE,
} from './help';

const _DEFAULT_SITE = 'default';
const JSONP_CALLBACK = 'callback=JSONP_CALLBACK';

// name of the SDK module
const _MODULE_NAME = 'WchSdk';
const _JSONP_CALLBACK_PREFIX = `${_MODULE_NAME}.jsonp.callback.`;

const RENDERING_PREFIX = 'delivery/v1/rendering/';
const CONTEXT_PREFIX = 'delivery/v1/rendering/context/';
const SITES_PREFIX = 'delivery/v1/rendering/sites/';

// prefetch filter
export type PrefetchFilter = (aLevel: number, aNode: SiteNode) => boolean;

function _getSiteUrl(aSite: SiteNode): string {
    const siteID = aSite.data.id || _DEFAULT_SITE;
    // replace
    return `${SITES_PREFIX}${encodeURIComponent(siteID)}`;
}

/**
 * Returns the URI path for the rendering context
 *
 * @param aID   the ID
 * @return the path
 */
function _getRenderingContextURL(aID: string): string {
    return `${CONTEXT_PREFIX}${encodeURIComponent(aID)}`;
}

function _getHash(aToken: string): string {
    return hashToIdentifier(hashString(longHash(), aToken));
}

function _getPreFetchURL(aUrl: string): string {
    // append callback
    let url = aUrl;
    if (url.indexOf('?') >= 0) {
        url += '&' + JSONP_CALLBACK;
    } else {
        url += '?' + JSONP_CALLBACK;
    }
    // construct the digest
    const digest = _getHash(url);
    // replace
    return url.replace(/=JSONP_CALLBACK(&|$)/,
        (match, suffix) => `=${encodeURIComponent(_JSONP_CALLBACK_PREFIX + digest)}${suffix}`);
}

interface Level {
    id: string;
    data: any;
    level: number;
}

function _forEachContent(aSite: SiteNode, aLevel: number, aFilter: PrefetchFilter): Observable<Level> {
    // check if we have a content item
    const contentId = (aSite && aSite.data && aSite.data.contentId) ? aSite.data.contentId : undefined;
    const data = (aSite && aSite.data) ? aSite.data : undefined;
    const obContentId: Observable<Level> = (contentId && aFilter(aLevel, aSite)) ? of({
        id: contentId,
        data: data,
        level: aLevel
    }) : empty();
    // children
    const children = (aSite && aSite.children) ? aSite.children : [];
    const obChildren = from(children).pipe(
        mergeMap(child => _forEachContent(child, aLevel + 1, aFilter))
    );
    // merge
    return merge(obContentId, obChildren);
}

function _sortByNumber(aLeft: number, aRight: number): number {
    return (aLeft < aRight) ? -1 : (aLeft > aRight) ? +1 : 0;
}

function _sortLevels(aLeft: Level, aRight: Level): number {
    // compare by level, then by position, then by name
    let cmp = _sortByNumber(aLeft.level, aRight.level);
    if (cmp === 0) {
        // by position
        const lPos = aLeft.data ? aLeft.data.position : 0;
        const rPos = aRight.data ? aRight.data.position : 0;
        cmp = _sortByNumber(lPos, rPos);
        if (cmp === 0) {
            // sort by ID
            return aLeft.id.localeCompare(aRight.id);
        }
    }
    // ok
    return cmp;
}

function _createUrl(aApiBase: string, aRelUrl: string, bRelative: boolean): string {
    // parse
    const url = `${aApiBase}${aRelUrl}`;
    const parsed = urlParse(url);
    return (bRelative ? parsed.path : `//${parsed.host}${parsed.path}`) || url;
}

function _getRenderingSuffix(aUrl: string): string {
    const idx = aUrl.indexOf(RENDERING_PREFIX);
    return aUrl.substr(idx);
}

function _getPreFetchContent(
    aApiBase: string,
    aSite: SiteNode | null | undefined,
    aFilter: PrefetchFilter,
    aLogger: Subject<string>): Observable<string> {
    // only continue if we have a site
    return _getPreFetchURLs(aApiBase, aSite, aFilter, aLogger, false).pipe(
        // strip the suffix
        map(_getRenderingSuffix)
    );
}

/**
 * Generates the prefetch URLs
 *
 * @param aApiBase  the API base URL
 * @param aSite     the site
 * @param aFilter
 * @param aLogger
 * @param bRelative
 */
function _getPreFetchURLs(aApiBase: string, aSite: SiteNode | null | undefined,
    aFilter: PrefetchFilter,
    aLogger: Subject<string>,
    bRelative: boolean): Observable<string> {
    // log this
    aLogger.next(`API base [${aApiBase}].`);
    // only continue if we have a site
    if (aSite) {
        // site URL
        const siteURL = _getSiteUrl(aSite);
        const sitePreFetchURL = _getPreFetchURL(siteURL);
        const obSitePreFetchURL = of(sitePreFetchURL);

        // generate a list
        const obChildren = _forEachContent(aSite, 0, aFilter).pipe(
            toArray(),
            map(ch => ch.sort(_sortLevels)),
            tap(levels => aLogger.next(`Levels [${JSON.stringify(levels, undefined, 2)}].`)),
            mergeMap(ch => from(ch)),
            map(l => l.id),
            map(id => _getRenderingContextURL(id)),
            map(_getPreFetchURL)
        );

        // ok
        return merge(obSitePreFetchURL, obChildren).pipe(
            map(relURL => _createUrl(aApiBase, relURL, bRelative))
        );
    } else {
        // log this
        aLogger.next('Site not available, skipping generation of preview URLs.');
        // nothing
        return empty();
    }
}

function _getPreFetchTemplate(): Observable<HandlebarsTemplateDelegate> {
    return hbsCompile(join(__dirname, '..', '..', '..', 'assets', 'prefetch.hbs'));
}

function _getPreFetchScriptTemplate(): Observable<HandlebarsTemplateDelegate> {
    return hbsCompile(join(__dirname, '..', '..', '..', 'assets', 'prefetch.script.hbs'));
}

function _getPreFetchScript(): Observable<string> {
    return readFile(join(__dirname, '..', '..', '..', 'assets', 'prefetch/add.prefetch.links.min.js'));
}

function _getPreFetchScriptMarkup(aContent: Observable<string>): Observable<string> {
    // returns the markup
    const rxTemplate = _getPreFetchScriptTemplate();
    const rxScript = _getPreFetchScript();
    // build the context
    const rxContent = aContent.pipe(
        toArray(),
        map(data => JSON.stringify(data))
    );
    const rxContext = combineLatest(rxScript, rxContent).pipe(
        map(([aScript, aContentLinks]) => ({
            SCRIPT: aScript,
            CONTENT: aContentLinks
        }))
    );
    // apply the template
    // convert
    return combineLatest(rxTemplate, rxContext).pipe(
        map(([tmp, ctx]) => tmp(ctx)));
}

function _createPreFetchScriptMarkup(aApiBase: string, aSite: SiteNode | null | undefined,
    aFilter: PrefetchFilter,
    aLogger: Subject<string>,
    bRelative: boolean): Observable<string> {
    // get the links
    const rxContent = _getPreFetchContent(aApiBase, aSite, aFilter, aLogger);
    // combine
    return _getPreFetchScriptMarkup(rxContent);
}

function _getPreFetchMarkup(aLinks: Observable<string>): Observable<string> {
    // returns the markup
    const rxTemplate = _getPreFetchTemplate();
    // convert all string into an array
    const rxContext = aLinks.pipe(toArray());
    // convert
    return combineLatest(rxTemplate, rxContext).pipe(
        map(([tmp, ctx]) => tmp(ctx))
    );
}

function _allNodes(aLevel: number, aNode: SiteNode): boolean {
    return true;
}

const REPLACE_REGEXP = /<!--\s+Prefetch begin\s+-->((.|\n|\r)*)<!--\s+Prefetch end\s+-->/g;
const BASE_REGEXP = /<base\s+[^>]*>/g;

function _insertMarkup(aPreRender: string, aMarkup: string): string {
    // check if we have an expression
    if (REPLACE_REGEXP.test(aMarkup)) {
        // replace
        return aMarkup.replace(REPLACE_REGEXP, () => aPreRender.trim());
    } else {
        // insert after the base tag
        return aMarkup.replace(BASE_REGEXP, (m, o) => m + '\n' + aPreRender.trim());
    }
}

function _rewriteIndex(aMarkup: string, aSrcDir: string): Observable<string> {
    // read the index file
    const fileName = join(aSrcDir, 'index.html');
    return readFile(fileName).pipe(
        map(content => _insertMarkup(aMarkup, content)),
        mergeMap(newMarkup => writeFile(fileName, newMarkup))
    );
}

/**
 * Generates layout and layout mappings
 *
 * add prefetch --data DATA_FOLDER --src SRC_FOLDER
 */
export class AddPreloadCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'prefetch';

    static readonly COMMAND = CMD_ADD;

    constructor(private options: Options) {
        super(options);
    }

    printHelp(): void {
        // log the command line
        console.log(AddPreloadCommand.COMMAND, AddPreloadCommand.TYPE,
            `--${OPTION_SOURCE} <SOURCE>`,
            `--${OPTION_DATA} <DATA>`,
            `[--${OPTION_NAME} <NAME>]`,
            `[--${OPTION_ROOT} <ROOT>]`,
            `[--${OPTION_RELATIVE}]`,
            `[--${OPTION_SCRIPT}]`,
            `[--${OPTION_VERBOSE}]`);
        // log the parameters
        console.log(`--${OPTION_SOURCE} <SOURCE>`, ':', HELP_COPY_SOURCE);
        console.log(`--${OPTION_DATA} <DATA>`, ':', HELP_DATA_DIRECTORY);
        console.log(`--${OPTION_NAME} <NAME>`, ':', HELP_APPLICATION_NAME);
        console.log(`--${OPTION_ROOT}`, ':', HELP_CONTEXT_ROOT);
        console.log(`--${OPTION_RELATIVE}`, ':', HELP_RELATIVE);
        console.log(`--${OPTION_VERBOSE}`, ':', HELP_VERBOSE);
    }

    process(): Observable<string> {
        // options
        const opts = this.options;
        // the directories
        const dataDirOpt = this.getDataDir();
        const appName = opts.name || this.getAppName();
        const contextRoot = `/${configTrimSlashes(opts.root || this.getContextRoot(appName))}`;
        // the directories
        const srcDirOpt = optNormalizeDir(firstOf(opts.src) || configIndexDirFromContextRoot(contextRoot, dataDirOpt));
        const srcDir = this.rxMkdirp(srcDirOpt);
        // base URL
        const apiBase = this.getApiBase();
        // relative or absolute
        const bRelative = !!opts.relative;
        // filter
        const filter: PrefetchFilter = _allNodes;

        // the logger
        const logger = this.getLogger();

        // log the settings
        logger.next(`Data dir [${dataDirOpt}], source dir [${srcDirOpt}], relative [${bRelative}].`);

        // tools
        const rxWchTools = this.getWchTools();

        // check the script case
        const bScript = !!opts.script;

        // get the markup
        const rxMarkup = combineLatest(rxWchTools, apiBase).pipe(
            mergeMap(([aWchTools, aApiBase]) =>
                bScript ? _createPreFetchScriptMarkup(aApiBase,
                    aWchTools.sites ? aWchTools.sites['default'] : undefined,
                    filter, logger, bRelative) :
                    _getPreFetchMarkup(
                        _getPreFetchURLs(aApiBase, aWchTools.sites ? aWchTools.sites['default'] : undefined,
                            filter, logger, bRelative)
                    )
            ));

        // rewrite the index
        const rxIndex = combineLatest(srcDir, rxMarkup).pipe(
            mergeMap(([aSrcDir, aMarkup]) => _rewriteIndex(aMarkup, aSrcDir))
        );

        // done
        return rxIndex;
    }
}
