/* Copyright IBM Corp. 2017 */
import * as Promise from 'bluebird';
import { spawn } from 'child_process';
import { replace as lodashReplace } from 'lodash';
import { join, normalize, relative } from 'path';
import { chdir, cwd, env } from 'process';
import { combineLatest, concat, defer, empty, from, merge, Observable, Observer, of, Subject } from 'rxjs';
import { distinct, finalize, last, map, mergeMap, mergeMapTo, shareReplay, take, tap } from 'rxjs/operators';
import { Writable } from 'stream';
import { parse as parseURL } from 'url';

import { firstOf } from '../utils/file.utils';
import { writeJsonFileSafe, writeTextFileSafe } from '../utils/rx.file';
import { hbsCompile } from '../utils/rx.hbs';
import { ensureTrailingSlash } from '../utils/url.utils';
import {
    APP_MODULE_FILE,
    CMD_CREATE,
    OPTION_DATA,
    OPTION_EXCLUDE_PATTERN,
    OPTION_INCLUDE_PATTERN,
    OPTION_NAME,
    OPTION_OVERRIDE,
    OPTION_PASSWORD,
    OPTION_RELATIVE,
    OPTION_ROOT,
    OPTION_SCSS,
    OPTION_SKIP_PULL,
    OPTION_SOURCE,
    OPTION_TYPE,
    OPTION_URL,
    OPTION_USER,
    OPTION_VERBOSE,
} from './../constants';
import { Options, optNormalizeDir } from './../options';
import { assertNotNull } from './../utils/assert';
import { canonicalizeJSON } from './../utils/json';
import { rxDeleteDir } from './../utils/rx.delete';
import { readFile, rxCopyFile, rxDeleteFile, rxMkdirp, rxValidateNotExists, walkFiles, writeFile } from './../utils/rx.walk';
import { kebabCase } from './../utils/transliteration';
import {
    FILE_WCHTOOLS_JSON,
    IdMapping,
    Matcher,
    readWchTools,
    wchDeliveryUrlFromApiUrl,
    wchGetIndexUrl,
    WchTools,
    wchToolsGetMatcher,
    WchToolsOptions,
} from './../utils/wchtools';
import { AbstractCommand } from './abstract.command';
import { AddBaseCommand } from './add.base.command';
import { AddEsiPolyfillCommand } from './add.esi.polyfill.command';
import { AddPreloadCommand } from './add.preload.command';
import { AddWebAppManifestCommand } from './add.webappmanifest.command';
import { ApplicationPullCommand } from './application.pull.command';
import { ApplicationPushCommand } from './application.push.command';
import { BuildDevelopmentCommand } from './build.development.command';
import { BuildProductionCommand } from './build.production.command';
import { Command } from './command';
import { CompressAssetsCommand } from './compress.assets.command';
import { CopyBuildCommand } from './copy.build.command';
import { CreateComponentsCommand } from './create.components.command';
import { CreateLayoutsCommand } from './create.layouts.command';
import { CreatePageTypeCommand } from './create.page.type.command';
import { CreateSiteCommand } from './create.site.command';
import { CreateWebAppManifestCommand } from './create.webappmanifest.command';
import { HELP_PROJECT_OVERRIDE, HELP_PROJECT_SOURCE, HELP_SKIP_PULL, HELP_VERBOSE } from './help';
import { NormalizeJsonCommand } from './normalize.json.command';
import { RunWchToolsCommand } from './run.wchtools.command';
import { ShowTypesCommand } from './show.types.command';
import { ShowUrlCommand } from './show.url.command';

const cli = require('@angular/cli').default;

const notice = /create\s+([^\s]+)/g;
const noColors = /\x1B\[([0-9]{1,2}(;[0-9]{1,2})?)?[mGK]/g;

const ASSETS_BASE = join(__dirname, '..', '..', '..', 'assets');

function _isInternalHost(aBaseURL: string): boolean {
    const url = parseURL(aBaseURL);
    return !!(url.host && !url.host.startsWith('www.') && (url.host.indexOf('digitalexperience.ibm.com') <= 0));
}

function _pullData(aDataDir: string,
    aConfig: WchToolsOptions,
    aOptions: Options,
    aLogger: Subject<string>): Observable<string> {
    // build the command line
    const args = [
        normalize(join(__dirname, '../utils/fork.wchtools')),
        'pull',
        '--dir',
        aDataDir,
        '--user',
        aConfig.username,
        '--password',
        aConfig.password,
        '--url',
        aConfig.baseUrl,
        '-tlmsp'];
    // verbose flag
    if (aOptions.verbose) {
        args.push('--verbose');
    }

    // execute
    return Observable.create((o: Observer<string>) => {

        // spawn the process
        const childProcess = spawn(process.execPath, args, {
            'cwd': aDataDir,
            'env': Object.assign(
                {
                    'NODE_TLS_REJECT_UNAUTHORIZED': '0'
                }, env
            )
        });
        // start
        o.next('start');
        // attach
        childProcess.stdout.on('data', data => aLogger.next(data.toString().replace(noColors, '')));
        childProcess.stderr.on('data', data => aLogger.next(data.toString().replace(noColors, '')));
        // end
        childProcess.on('close', code => o.complete());
        childProcess.on('error', err => o.error(err));
    });
}

class OutputParser extends Writable {

    files = new Subject<string>();

    constructor(private aBaseDir: string) {
        super();
    }

    _write(chunk: any, encoding: string, callback: Function) {

        // extract
        const strg = chunk.toString().replace(noColors, '').trim();
        const res = notice.exec(strg);
        if (res && (res.length > 1)) {
            // dump
            this.files.next(join(this.aBaseDir, res[1]));
        }
        callback();
    }

    _final(callback: Function) {
        console.log('done');
        super._final(callback);
    }
}

function _pullTypes(aSrcDir: string, aName: string,
    aConfig: WchToolsOptions, aOptions: Options, aLogger: Subject<string>): Observable<string> {
    // just pull once
    const dataDir = join(aSrcDir, aName, 'data');
    // log this
    aLogger.next(`Pulling types to [${dataDir}].`);
    // continue
    return rxMkdirp(dataDir).pipe(
        mergeMap(dir => _pullData(dir, aConfig, aOptions, aLogger).pipe(
            last(),
            mergeMap(() => walkFiles(dir).pipe(map(desc => desc.absPath))))
        ), shareReplay(1));
}

function _createLayoutsAndMappings(aSrcDir: string, aName: string, aOptions: Options,
    aLogger: Subject<string>): Observable<string> {
    // the data dir
    const dataDir = join(aSrcDir, aName, 'data');
    // execute the layouts command
    const opt: Options = {
        command: CreateLayoutsCommand.COMMAND,
        commandType: CreateLayoutsCommand.TYPE,
        data: dataDir,
        verbose: aOptions.verbose,
        override: aOptions.override,
        include: aOptions.include,
        exclude: aOptions.exclude,
        allArgs: aOptions.allArgs
    };
    // create
    return of(opt).pipe(
        map(o => new CreateLayoutsCommand(opt)),
        mergeMap(cmd => cmd.process()),
        shareReplay());
}

function _createComponents(aSrcDir: string, aName: string, aOptions: Options,
    aLogger: Subject<string>): Observable<string> {
    // the data dir
    const dataDir = join(aSrcDir, aName, 'data');
    const srcDir = join(aSrcDir, aName, 'src', 'app');
    // execute the layouts command
    const opt: Options = {
        command: CreateComponentsCommand.COMMAND,
        commandType: CreateComponentsCommand.TYPE,
        data: dataDir,
        src: srcDir,
        verbose: aOptions.verbose,
        override: aOptions.override,
        include: aOptions.include,
        exclude: aOptions.exclude,
        editable: aOptions.editable,
        skipRegistration: true,
        scss: true,
        allArgs: aOptions.allArgs
    };
    // process
    return of(opt).pipe(
        map(o => new CreateComponentsCommand(opt)),
        mergeMap(cmd => cmd.process()),
        shareReplay()
    );
}

function _createNgCli(aSrcDir: string, aName: string, aLogger: Subject<string>): Observable<string> {
    // sanity check
    assertNotNull(aSrcDir, 'srcDir');
    assertNotNull(aName, 'name');
    // trigger
    return defer(() => of(aSrcDir).pipe(
        mergeMap(srcDir => {
            // notify the start
            aLogger.next('createNgCli start');
            // parse the command output
            const outParser = new OutputParser(normalize(srcDir));
            // the files handler
            const files = outParser.files;

            // log the arguments
            const cliArgs = ['new', aName, '--skip-install', '--skip-git', '--skip-commit', '--routing', '--style', 'scss'];
            // notify the start
            aLogger.next(`Calling CLI with ${JSON.stringify(cliArgs)}`);

            // execute the command, we will sync on files
            Promise.resolve(cli({
                cliArgs: cliArgs,
                directory: srcDir,
                inputStream: process.stdin,
                outputStream: outParser
            }))
                .then(() => files.complete())
                .catch(err => files.error(err));
            // returns the generated files
            return files.asObservable();
        }),
        finalize(() => aLogger.next('createNgCli end')),
        shareReplay())
    );
}

function _log(aData: string) {
    console.log('CreateProjectCommand', aData);
}

function _createAppConfig(aSourceDir: string, aDataDir: string, aOverride: boolean): Observable<string> {

    // compute the relative path
    const dstDir = join(aSourceDir, 'src', 'app');
    const dstFile = join(dstDir, 'app.config.ts');
    const relPath = lodashReplace(relative(dstDir, aDataDir), /\\/g, '/');

    // prepare the hbs context
    const ctx = {
        dataDir: `./${relPath}/`
    };

    // load the template
    const rxTemplate = hbsCompile(join(ASSETS_BASE, 'app.config.ts.hbs'));
    // transform
    return rxTemplate.pipe(map(tmp => tmp(ctx)),
        mergeMap(content => writeTextFileSafe(dstFile, content, aOverride)));
}

function _fixAppModule(aSource: string): string {
    // insert the layouts import
    let src =
        `import { LAYOUTS } from './layouts';
import { WchNgModule } from '@ibm-wch-sdk/ng';
import { environment } from '../environments/environment';
${aSource}`;
    // add the module to the end of the declarations
    src = src.replace(/(declarations:\s+\[)([^\]]+)(\])/g, '$1$2, ...LAYOUTS$3');
    // add the entry components
    src = src.replace(/(bootstrap:)/g, 'entryComponents: [...LAYOUTS],\n$1');
    // add the SDK
    src = src.replace(/(imports:\s+\[)([^\]]+)(\])/g, '$1$2, WchNgModule.forRoot(environment)$3');

    return src;
}

function _rewriteAppModule(aSrcDir: string, aName: string, aLogger: Subject<string>): Observable<string> {
    // name
    const name = join(aSrcDir, aName, 'src', 'app', APP_MODULE_FILE);
    return readFile(name).pipe(
        map(_fixAppModule),
        mergeMap(data => writeFile(name, data))
    );
}

const envRegExp = /(environment\s+=\s+\{)([^\}]+)(\})/g;

function _fixDevEnvironment(aSource: string, aApiUrl: string): string {
    // insert the layouts import
    let src = aSource;
    // prefix
    const prefix = 'import { apiUrl } from \'./../app/app.config\';\n';
    // fragment
    const fragment = `, apiUrl
, httpOptions: { pollTime: 20 * 1000 }, \
httpPreviewOptions: { pollTime: 10 * 1000 }`;
    // add the delivery and api URL
    src = prefix + src.replace(envRegExp, '$1$2' + fragment + '$3');

    return src;
}

function _fixProdEnvironment(aSource: string, aApiUrl: string): string {
    // insert the layouts import
    let src = aSource;
    // deduce the delivery URL from the api URL
    const deliveryUrl = wchDeliveryUrlFromApiUrl(aApiUrl);
    // fragment
    const fragment = `, httpOptions: { pollTime: 5 * 60 * 1000 }, \
httpPreviewOptions: { pollTime: 10 * 1000 }`;
    // add the delivery and api URL
    src = src.replace(envRegExp, '$1$2' + fragment + '$3');

    return src;
}

function _rewriteEnvironment(aSrcDir: string, aName: string, aApiUrl: string, aLogger: Subject<string>): Observable<string> {
    // dev env
    const devName = join(aSrcDir, aName, 'src', 'environments', 'environment.ts');
    const dev = readFile(devName).pipe(
        map(data => _fixDevEnvironment(data, aApiUrl)),
        mergeMap(data => writeFile(devName, data))
    );

    // dev env
    const prodEnv = join(aSrcDir, aName, 'src', 'environments', 'environment.prod.ts');
    const prod = readFile(prodEnv).pipe(
        map(data => _fixProdEnvironment(data, aApiUrl)),
        mergeMap(data => writeFile(prodEnv, data))
    );

    // merge
    return merge(dev, prod);
}

function _fixRoutingModule(aSource: string, aTools: WchTools): string {
    // find the first page
    let defaultRoute = '/home/';
    if (aTools && aTools.sites && aTools.sites.default) {
        // access the default site
        const defSite = aTools.sites.default;
        if (defSite.children && (defSite.children.length > 0)) {
            // use the first child as default, do not encode this (strange!)
            defaultRoute = `/${defSite.children[0].data.segment}/`;
        }
    }

    // insert the layouts import
    let src = `import { PageComponent } from '@ibm-wch-sdk/ng';\n${aSource}`;
    // add the module to the end of the declarations
    const replace = `{
        path: '',
        redirectTo: '${defaultRoute}',
        pathMatch: 'full'
      },
      {
        path: '**',
        component: PageComponent
      }`;
    src = src.replace(/(const\s+routes:\s+Routes\s+=\s+\[)([^;]*)(\];)/g, '$1' + replace + '$3');

    // use hash based navigation
    const hashRegexp = /(RouterModule\.forRoot\s*\()([^\)]*)(\))/g;
    src = src.replace(hashRegexp, '$1$2, { useHash: true }$3');

    return src;
}

function _rewriteRoutingModule(aSrcDir: string, aName: string,
    aLogger: Subject<string>): Observable<string> {
    // data directory
    const dataDir = join(aSrcDir, aName, 'data');
    // name
    const name = join(aSrcDir, aName, 'src', 'app', 'app-routing.module.ts');
    return readFile(name).pipe(
        mergeMap(code => readWchTools(dataDir).pipe(map(tools => _fixRoutingModule(code, tools)))),
        mergeMap(data => writeFile(name, data))
    );
}

function _fixIndex(aSource: string): string {
    // insert the layouts import
    let src = aSource;
    // add the module to the end of the declarations
    src = src.replace(/(<base\s+href=\")([^\"]*)(\">)/g, '$1.$3');

    return src;
}

function _rewriteIndex(aSrcDir: string, aName: string, aLogger: Subject<string>): Observable<string> {
    // name
    const name = join(aSrcDir, aName, 'src', 'index.html');
    return readFile(name).pipe(
        map(_fixIndex),
        mergeMap(data => writeFile(name, data))
    );
}

function _fixLint(aLint: any): any {
    // need this due to still unresolved https://github.com/mgechev/codelyzer/issues/191
    const rules = aLint.rules || (aLint.rules = {});
    rules['no-access-missing-member'] = false;
    // SDK uses 'on' as prefix
    rules['no-output-on-prefix'] = false;
    // add rxjs rules
    const ext = aLint.extends || (aLint.extends = []);
    ext.push('rxjs-tslint-rules');
    // add extra rules
    rules['rxjs-add'] = { 'severity': 'error' };
    rules['rxjs-no-unused-add'] = { 'severity': 'error' };
    rules['rxjs-no-subject-unsubscribe'] = { 'severity': 'error' };
    // ok
    return aLint;
}

function _rewriteLint(aSrcDir: string, aName: string, aLogger: Subject<string>): Observable<string> {
    // name
    const name = join(aSrcDir, aName, 'tslint.json');
    return readFile(name).pipe(
        map(data => JSON.parse(data)),
        map(_fixLint),
        map(canonicalizeJSON),
        map(data => JSON.stringify(data, undefined, 2)),
        mergeMap(data => writeFile(name, data))
    );
}

function _fixCli(aCli: any): any {
    // add the flag to all fields
    const apps = aCli.apps || (aCli.apps = []);
    // iterate and add
    apps.forEach((app: any) => app.serviceWorker = false);
    // ok
    return aCli;
}

function _rewriteCli(aSrcDir: string, aName: string, aLogger: Subject<string>): Observable<string> {
    // name
    const name = join(aSrcDir, aName, '.angular-cli.json');
    return readFile(name).pipe(
        map(data => JSON.parse(data)),
        map(_fixCli),
        map(canonicalizeJSON),
        map(data => JSON.stringify(data, undefined, 2)),
        mergeMap(data => writeFile(name, data))
    );
}

function _createWchtoolsOptions(aDataDir: string, aConfig: WchToolsOptions, aLogger: Subject<string>): Observable<string> {
    // build the object
    const opts = {
        'username': aConfig.username,
        'x-ibm-dx-tenant-base-url': aConfig.baseUrl,
        'ignore_additive': true
    };
    // serialize and write
    const name = join(aDataDir, FILE_WCHTOOLS_JSON);
    // construct the data directory
    return rxMkdirp(aDataDir).pipe(
        mergeMapTo(writeFile(name, JSON.stringify(canonicalizeJSON(opts), undefined, 2)))
    );
}

function _getDataDir(aSrcDir: string, aName: string): string {
    return join(aSrcDir, aName, 'data');
}

function _updatePackage(aPackage: any, aName: string, bEditable: boolean): any {
    // add the dependencies
    const dependencies = aPackage.dependencies;
    const devDependencies = aPackage.devDependencies;
    // add service workers
    if (!devDependencies['@angular/service-worker']) {
        devDependencies['@angular/service-worker'] = '^6.0.0';
    }
    // SDK dependencies
    dependencies['@ibm-wch-sdk/ng'] = '^6.0.0';
    devDependencies['@ibm-wch-sdk/cli'] = '^6.0.0';
    // add dependency if editable
    if (bEditable) {
        dependencies['@ibm-wch-sdk/ng-edit'] = '^6.0.0';
    }
    // some RX stuff
    devDependencies['rxjs-tslint-rules'] = '^4.1.1';
    devDependencies['cpx'] = '^1.5.0';
    devDependencies['rimraf'] = '^2.6.2';
    devDependencies['npm-run-all'] = '^4.1.3';
    // engines
    aPackage.engines = {
        'node': '>= 7.0',
        'npm': '>= 5.0'
    };
    // config
    const cfg: any = aPackage.config = {};
    cfg[OPTION_DATA] = './data';
    cfg[OPTION_SOURCE] = './src';
    cfg[OPTION_NAME] = aName;
    cfg[OPTION_ROOT] = '/';
    const CLI = 'ibm-wch-sdk-cli';

    // add scripts
    const scripts = aPackage.scripts;
    scripts['build:production'] = `${CLI} \
${BuildProductionCommand.COMMAND} \
${BuildProductionCommand.TYPE} \
--${OPTION_RELATIVE}`;

    // add scripts
    scripts['build:development'] = `${CLI} \
${BuildDevelopmentCommand.COMMAND} \
${BuildDevelopmentCommand.TYPE} \
--${OPTION_RELATIVE}`;

    scripts['run:wchtools'] = `${CLI} \
${RunWchToolsCommand.COMMAND} \
${RunWchToolsCommand.TYPE}`;

    scripts['gen:layouts'] = `${CLI} \
${CreateLayoutsCommand.COMMAND} \
${CreateLayoutsCommand.TYPE} \
--${OPTION_DATA} ${cfg[OPTION_DATA]}/`;

    scripts['gen:components'] = `${CLI} \
${CreateComponentsCommand.COMMAND} \
${CreateComponentsCommand.TYPE} \
--${OPTION_SOURCE} ${cfg[OPTION_SOURCE]}/app/ \
--${OPTION_DATA} ${cfg[OPTION_DATA]} \
--${OPTION_SCSS}`;

    scripts['package:copyResources'] = `${CLI} \
${CopyBuildCommand.COMMAND} \
${CopyBuildCommand.TYPE}`;

    scripts['package:addPrefetch'] = `${CLI} \
${AddPreloadCommand.COMMAND} \
${AddPreloadCommand.TYPE} \
--${OPTION_RELATIVE}`;

    scripts['package:addEsiPolyfill'] = `${CLI} \
${AddEsiPolyfillCommand.COMMAND} \
${AddEsiPolyfillCommand.TYPE}`;

    scripts['package:createSite'] = `${CLI} \
${CreateSiteCommand.COMMAND} \
${CreateSiteCommand.TYPE}`;

    scripts['package:compress'] = `${CLI} \
    ${CompressAssetsCommand.COMMAND} \
    ${CompressAssetsCommand.TYPE}`;

    scripts['package:production'] = `npm run package:copyResources \
&& npm run package:addPrefetch \
&& npm run package:addBaseUrl \
&& npm run package:addManifest \
&& npm run package:addEsiPolyfill \
&& npm run package:createSite`;

    scripts['deploy:push'] = `${CLI} \
${ApplicationPushCommand.COMMAND} \
${ApplicationPushCommand.TYPE} \
--${OPTION_DATA} ${cfg[OPTION_DATA]}/`;

    scripts['gen:pull'] = `${CLI} \
${ApplicationPullCommand.COMMAND} \
${ApplicationPullCommand.TYPE} \
--${OPTION_DATA} ${cfg[OPTION_DATA]}/`;

    scripts['gen:pagetype'] = `${CLI} \
${CreatePageTypeCommand.COMMAND} \
${CreatePageTypeCommand.TYPE} \
--${OPTION_SOURCE} ${cfg[OPTION_SOURCE]}/app/ \
--${OPTION_DATA} ${cfg[OPTION_DATA]} \
--${OPTION_SCSS} \
--${OPTION_TYPE}`;

    scripts['show:types'] = `${CLI} \
${ShowTypesCommand.COMMAND} \
${ShowTypesCommand.TYPE}`;

    scripts['show:url'] = `${CLI} \
${ShowUrlCommand.COMMAND} \
${ShowUrlCommand.TYPE}`;

    scripts['package:addBaseUrl'] = `${CLI} \
${AddBaseCommand.COMMAND} \
${AddBaseCommand.TYPE} \
--${OPTION_RELATIVE}`;

    scripts['package:addManifest'] = `${CLI} \
${AddWebAppManifestCommand.COMMAND} \
${AddWebAppManifestCommand.TYPE} \
--${OPTION_RELATIVE}`;

    scripts['gen:normalize'] = `${CLI} \
${NormalizeJsonCommand.COMMAND} \
${NormalizeJsonCommand.TYPE}`;

    scripts['deploy'] = 'npm run build:production && npm run package:production && npm run deploy:push && npm run show:url';
    scripts['lint'] = 'ng lint --type-check';

    return aPackage;
}

function _updatePackageJson(aSrcDir: string, aName: string, aEditable: boolean, aLogger: Subject<string>): Observable<string> {
    // load the package json
    const name = join(aSrcDir, aName, 'package.json');
    return readFile(name).pipe(
        map(data => JSON.parse(data)),
        map(pkg => _updatePackage(pkg, aName, aEditable)),
        map(canonicalizeJSON),
        map(json => JSON.stringify(json, null, 2)),
        mergeMap(pkg => writeFile(name, pkg))
    );
}

function _updateVisualStudioCodeSettings(aSrcDir: string, aName: string, aWchtoolsOptions: WchToolsOptions, aLogger: Subject<string>): Observable<string> {
    // templates
    const templateFile = join(ASSETS_BASE, 'vscode.settings.json.template');
    const rxTemplateContent = readFile(templateFile).pipe(map(data => JSON.parse(data)));
    // target path
    const dstDir = join(aSrcDir, aName, '.vscode');
    const dstFile = join(dstDir, 'settings.json');
    // construct target file
    const rxDir = rxMkdirp(dstDir);
    // rewrite
    const apiUrl = ensureTrailingSlash(aWchtoolsOptions.baseUrl);
    const userName = aWchtoolsOptions.username;
    // update
    return combineLatest(rxTemplateContent.pipe(map(data => Object.assign(data, { 'wch.username': userName, 'wch.apiUrl': apiUrl }))), rxDir).pipe(
        mergeMap(([data, dir]) => writeJsonFileSafe(dstFile, data, true))
    );
}

function _updateVisualStudioCodeLaunch(aSrcDir: string, aName: string, aWchtoolsOptions: WchToolsOptions, aLogger: Subject<string>): Observable<string> {
    // templates
    const templateFile = join(ASSETS_BASE, 'vscode.launch.json.template');
    const rxTemplateContent = readFile(templateFile).pipe(map(data => JSON.parse(data)));
    // target path
    const dstDir = join(aSrcDir, aName, '.vscode');
    const dstFile = join(dstDir, 'launch.json');
    // construct target file
    const rxDir = rxMkdirp(dstDir);
    // get the URLs
    const rxIndex = wchGetIndexUrl('/', aWchtoolsOptions);
    // update the data
    const rxData = combineLatest(rxTemplateContent, rxIndex).pipe(
        map(([data, index]) => {
            // update the data
            const cfg = data.configurations || (data.configurations = []);
            // insert a special server side launch config
            cfg.push({
                'name': 'Launch on WCH',
                'type': 'chrome',
                'request': 'launch',
                'url': index
            });
            // ok
            return data;
        }));
    // update
    return combineLatest(rxData, rxDir, rxIndex).pipe(
        mergeMap(([data, dir, index]) => writeJsonFileSafe(dstFile, data, true))
    );
}

function _updateVisualStudioCodeTasks(aSrcDir: string, aName: string, aWchtoolsOptions: WchToolsOptions, aLogger: Subject<string>): Observable<string> {
    // templates
    const templateFile = join(ASSETS_BASE, 'vscode.tasks.json.template');
    const rxTemplateContent = readFile(templateFile).pipe(map(data => JSON.parse(data)));
    // target path
    const dstDir = join(aSrcDir, aName, '.vscode');
    const dstFile = join(dstDir, 'tasks.json');
    // construct target file
    const rxDir = rxMkdirp(dstDir);
    // update the data
    const rxData = rxTemplateContent;
    // update
    return combineLatest(rxData, rxDir).pipe(
        mergeMap(([data, dir]) => writeJsonFileSafe(dstFile, data, true))
    );
}

function _updateProjectTypings(aTypingsFile: string, aLogger: Subject<string>): Observable<string> {
    // templates
    const templateFile = join(ASSETS_BASE, 'typings.d.ts.template');
    const rxTemplateContent = readFile(templateFile);

    const typingsFile = aTypingsFile;
    const rxTypingsContent = readFile(typingsFile);

    return combineLatest(rxTemplateContent, rxTypingsContent).pipe(
        map(([templateContent, typingsContent]) => `${typingsContent}\n${templateContent}`),
        mergeMap(data => writeFile(typingsFile, data))
    );
}

function _updateTypings(aSrcDir: string, aName: string, aLogger: Subject<string>): Observable<string> {

    const typingsFile = join(aSrcDir, aName, 'src', 'typings.d.ts');
    return _updateProjectTypings(typingsFile, aLogger);
}

function _copyResources(aSrcDir: string, aName: string, aOpts: WchToolsOptions, aLogger: Subject<string>): Observable<string> {
    // check if we need to copy the npmrc
    const res: Observable<string>[] = [];
    if (_isInternalHost(aOpts.baseUrl)) {
        // copy
        res.push(rxCopyFile(join(ASSETS_BASE, 'npmrc.template'), join(aSrcDir, aName, '.npmrc')));
        res.push(rxCopyFile(join(ASSETS_BASE, 'yarnrc.template'), join(aSrcDir, aName, '.yarnrc')));
    }
    res.push(rxCopyFile(join(ASSETS_BASE, 'gitignore.template'), join(aSrcDir, aName, '.gitignore')));
    res.push(rxCopyFile(join(ASSETS_BASE, 'npmignore.template'), join(aSrcDir, aName, '.npmignore')));
    // wait for all
    return merge(...res);
}

function _matches(aName: string, aMatcher: Matcher, aLogger: Subject<string>): boolean {
    // check
    const result = aMatcher(aName);
    aLogger.next(`Matching [${aName}] results in [${result}].`);
    return result;
}

declare type TypeMatcher = (aType: any) => boolean;

/**
 * Selects components by applying a matching function
 *
 * @param aMatcher  the matcher
 * @param aLogger   the logger
 */
function _byMatcher(aMatcher: Matcher, aLogger: Subject<string>): TypeMatcher {
    // execute the check
    return (aType: any) => aType && aType.name && _matches(aType.name, aMatcher, aLogger);
}

function _getFilesToDelete(aTools: WchTools, aMatcher: TypeMatcher): Observable<string> {
    // identify all valid types
    const validTypes: IdMapping<any> = {};
    Object.keys(aTools.types).map(id => aTools.types[id])
        .filter(type => aMatcher(type))
        .forEach(type => validTypes[type.id] = type);
    // check for the valid layout mappings
    const validMappings: IdMapping<any> = {};
    Object.keys(aTools.layoutMappings)
        .map(id => aTools.layoutMappings[id])
        .filter(mapping => validTypes[mapping.type.id])
        .forEach(mapping => validMappings[mapping.id] = mapping);
    // check for the valid layouts
    const validLayouts: IdMapping<any> = {};
    Object.keys(validMappings)
        .map(id => validMappings[id])
        .forEach(mapping => mapping.mappings.forEach((m: any) => m.layouts.forEach((l: any) => validLayouts[l.id] = l)));

    // the invalid types
    const typeFiles = Object.keys(aTools.types)
        .filter(id => !validTypes[id])
        .map(id => aTools.idToFile[id]);

    // the invalid mappings
    const mappingFiles = Object.keys(aTools.layoutMappings)
        .filter(id => !validMappings[id])
        .map(id => aTools.idToFile[id]);

    // the invalid layouts
    const layoutFiles = Object.keys(aTools.layouts)
        .filter(id => !validLayouts[id])
        .map(id => aTools.idToFile[id]);

    // returns the files to delete
    return merge(from(typeFiles), from(mappingFiles), from(layoutFiles));
}

function _filterWchToolsResources(aSrcDir: string,
    aName: string,
    aOptions: Options,
    aLogger: Subject<string>): Observable<string> {
    // construct the matcher
    const matcher: Matcher = wchToolsGetMatcher(aOptions);
    const typeMatcher: TypeMatcher = _byMatcher(matcher, aLogger);
    // scan the types
    const dir = join(aSrcDir, aName, 'data');
    return readWchTools(dir).pipe(
        mergeMap(tools => _getFilesToDelete(tools, typeMatcher)),
        mergeMap(rxDeleteFile)
    );
}

function _addManifest(aSrcDir: string,
    aName: string,
    aOptions: Options): Observable<string> {
    // source
    const srcDir = join(aSrcDir, aName);
    const dataDir = join(aSrcDir, aName, 'data');
    // build the options
    return of({
        command: CreateWebAppManifestCommand.COMMAND,
        commandType: CreateWebAppManifestCommand.TYPE,
        data: dataDir,
        src: srcDir,
        name: aName,
        root: aName,
        verbose: aOptions.verbose,
        allArgs: aOptions.allArgs
    }).pipe(
        map(opt => new CreateWebAppManifestCommand(opt)),
        mergeMap(cmd => cmd.process())
    );
}

/**
 * Generates angular components based on type and layout mapping definitions
 *
 * create project --src SOURCE_CODE_FOLDER --name NAME
 */
export class CreateProjectCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'project';

    static readonly COMMAND = CMD_CREATE;

    constructor(private options: Options) {
        super(options);
    }

    printHelp(): void {
        // log the command line
        console.log(CreateProjectCommand.COMMAND, CreateProjectCommand.TYPE,
            `--${OPTION_SOURCE} <SOURCE>`,
            `--${OPTION_NAME} <NAME>`,
            `[--${OPTION_URL} <URL>]`,
            `[--${OPTION_USER} <USER>]`,
            `[--${OPTION_PASSWORD} <PASSWORD>]`,
            `[--${OPTION_INCLUDE_PATTERN} <INCLUDE_PATTERN>]`,
            `[--${OPTION_EXCLUDE_PATTERN} <EXCLUDE_PATTERN>]`,
            `[--${OPTION_SKIP_PULL}]`,
            `[--${OPTION_OVERRIDE}]`,
            `[--${OPTION_VERBOSE}]`);
        // log the parameters
        console.log(`--${OPTION_SOURCE} <SOURCE>`, ':', HELP_PROJECT_SOURCE);
        console.log(`--${OPTION_SKIP_PULL}`, ':', HELP_SKIP_PULL);
        console.log(`--${OPTION_OVERRIDE}`, ':', HELP_PROJECT_OVERRIDE);
        console.log(`--${OPTION_VERBOSE}`, ':', HELP_VERBOSE);
    }

    process(): Observable<string> {
        // options
        const opts = this.options;
        // switch to the target directory
        const rxSrcDir = this.rxMkdirp(optNormalizeDir(firstOf(this.options.src) || '.'));
        // capture the current working directory
        const currentDir = cwd();
        const name = kebabCase(opts.name || this.getAppName());
        // pull content?
        const bSkipPull = opts.skipPull || false;
        // check if we would override
        const rxClean = (this.options.override && name)
            ? rxSrcDir.pipe(mergeMap(aSrcDir => rxDeleteDir(join(aSrcDir, name)).pipe(mergeMap(() => rxSrcDir))))
            : rxSrcDir.pipe(mergeMap(aSrcDir => rxValidateNotExists(join(aSrcDir, name)).pipe(mergeMap(() => rxSrcDir))));

        // logger
        const logger = new Subject<string>();
        // add some logging
        if (this.options.verbose) {
            logger.subscribe(_log);
        }

        // base URL
        const toolsOpts = this.getWchToolsOptions().pipe(shareReplay(1));

        // check if we create an inline edit experience
        const bEditable = opts.editable || false;

        // construct the target
        return rxClean.pipe(
            tap(chdir),
            mergeMap(aSrcDir => {
                // url
                const apiUrl = this.options.url || '';
                // merge the different steps
                const app = _createNgCli(aSrcDir, name, logger);
                const data = bSkipPull ? app : combineLatest(toolsOpts, app.pipe(take(1))).pipe(
                    mergeMap(([aToolOpts]) => concat(
                        bSkipPull ? empty() : _pullTypes(aSrcDir, name, aToolOpts, this.options, logger),
                        _filterWchToolsResources(aSrcDir, name, this.options, logger))));
                const copyResources = toolsOpts.pipe(mergeMap(aToolsOpts =>
                    _copyResources(aSrcDir, name, aToolsOpts, logger)));
                const wchToolsOptions = toolsOpts.pipe(mergeMap(aToolsOpts =>
                    _createWchtoolsOptions(_getDataDir(aSrcDir, name), aToolsOpts, logger)));
                const appConfig = _createAppConfig(join(aSrcDir, name), _getDataDir(aSrcDir, name), !!opts.override);
                const layouts = _createLayoutsAndMappings(aSrcDir, name, this.options, logger);
                const components = _createComponents(aSrcDir, name, this.options, logger);
                const pkg = _updatePackageJson(aSrcDir, name, bEditable, logger);
                const typings = _updateTypings(aSrcDir, name, logger);
                const vcSettings = toolsOpts.pipe(mergeMap(aToolsOpts => _updateVisualStudioCodeSettings(aSrcDir, name, aToolsOpts, logger)));
                const vcLaunch = toolsOpts.pipe(mergeMap(aToolsOpts => _updateVisualStudioCodeLaunch(aSrcDir, name, aToolsOpts, logger)));
                const vcTasks = toolsOpts.pipe(mergeMap(aToolsOpts => _updateVisualStudioCodeTasks(aSrcDir, name, aToolsOpts, logger)));
                const appModule = _rewriteAppModule(aSrcDir, name, logger);
                const routingModule = _rewriteRoutingModule(aSrcDir, name, logger);
                const environment = _rewriteEnvironment(aSrcDir, name, apiUrl, logger);
                const index = _rewriteIndex(aSrcDir, name, logger);
                const lint = _rewriteLint(aSrcDir, name, logger);
                const ngCli = _rewriteCli(aSrcDir, name, logger);
                const manifest = _addManifest(aSrcDir, name, this.options);

                // files that can be run in parallel
                const rxFiles = merge(typings, vcLaunch, vcTasks, vcSettings, wchToolsOptions, appConfig);

                // merge all of these
                return merge(app, concat(data, layouts, components, pkg, rxFiles, appModule,
                    routingModule, environment, index, lint, manifest, ngCli, copyResources));
            }),
            distinct(),
            finalize(() => chdir(currentDir))
        );
    }
}

export {
    _createNgCli as createNgCli,
    _copyResources as createProjectCopyResources,
    _createWchtoolsOptions as createProjectCreateWchtoolsOptions,
    _updateProjectTypings as createProjectUpdateTypings,
    _createAppConfig as createProjectCreateAppConfig
};
