/* Copyright IBM Corp. 2017 */
import { createHash } from 'crypto';
import { createWriteStream } from 'graceful-fs';
import { escape as htmlEscape, isArray, isString } from 'lodash';
import { normalize, resolve } from 'path';
import { Observable, of } from 'rxjs';
import { mergeMap } from 'rxjs/operators';
import { Writable } from 'stream';

import { CMD_SHOW } from './../constants';
import { Options } from './../options';
import { SiteNode, WchTools } from './../utils/wchtools';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';

const nodeMap: { [id: string]: string } = {};

const colorMap: { [id: string]: string } = {};

let nodeCounter = 0;

function _getColorValue() {
    return Math.floor(Math.random() * 256);
}

function _getTypeColor(aID: string): string {
    // some random value
    let c = colorMap[aID];
    if (!c) {
        c = colorMap[aID] = '#' + createHash('sha256').update(aID).digest('hex').substr(0, 6);
    }
    return c;
}

function _getNodeName(aID: string): string {
    // test
    let name = nodeMap[aID];
    if (!name) {
        name = nodeMap[aID] = `node${nodeCounter++}`;
    }
    return name;
}

function _getLabel(aString: string): string {
    return htmlEscape(aString).replace(/\s/g, '\\n');
}

const SITE_TYPE = 'site';

function _writeContent(aContent: any, aTools: WchTools, aOut: Writable) {
    // write the content node
    const id = aContent.id;
    const name = _getNodeName(id);
    // elements
    const elements = aContent.elements;
    // write the site node
    aOut.write(`${name}[tooltip="${id}";label="${_getLabel(aContent.name)} | {`);
    // write the elements
    if (elements) {
        // write the node name
        Object.keys(elements).forEach((key: any, idx) => {
            // name the element
            const nodeID = _getNodeName(`${id}#${key})`);
            // write
            const value = elements[key];
            const sep = (idx > 0) ? '|' : '';
            aOut.write(`${sep}<${nodeID}>${htmlEscape(key)}`);
        });
    }
    // write
    aOut.write(`}", color="${_getTypeColor(aContent.typeId)}"];\n`);
    // write outgoing links
    if (elements) {
        // write the node name
        Object.keys(elements).forEach((key: any, idx) => {
            // name the element
            const nodeID = _getNodeName(`${id}#${key})`);
            const value = elements[key];
            // check if this is a reference
            if (value.elementType === 'reference') {
                // handle the values
                if (value.values && isArray(value.values)) {
                    // iterate over the values
                    value.values.forEach((dstLink: any) => {
                        // write the link
                        const dstName = _getNodeName(dstLink.id);
                        aOut.write(`"${name}":${nodeID} -> "${dstName}";\n`);
                    });
                }
                // handle a single value
                if (value.value && value.value.id && isString(value.value.id)) {
                    // write the link
                    const dstName = _getNodeName(value.value.id);
                    aOut.write(`"${name}":${nodeID} -> "${dstName}";\n`);
                }
            }
        });
    }
}

function _writeSite(aSite: SiteNode, aTools: WchTools, aOut: Writable) {
    // write the site
    const id = aSite.data.id;
    const name = _getNodeName(id);
    // write the site node
    aOut.write(`${name}[tooltip="${id}";label="${_getLabel(aSite.data.name)}", \
shape=oval, color="${_getTypeColor(aSite.data.classification)}"];\n`);
    // write the actual nodes
    if (aSite.children) {
        // iterate over the children
        aSite.children.forEach(site => {
            // write the site
            _writeSite(site, aTools, aOut);
            // writes the page links
            const dstName = _getNodeName(site.data.id);
            aOut.write(`"${name}" -> "${dstName}";\n`);
        });
    }
    // check of we have content
    const contentId = aSite.data.contentId;
    if (contentId) {
        // write the link
        const dstName = _getNodeName(contentId);
        aOut.write(`"${name}" -> "${dstName}";\n`);
    }
}

function _writeDigraph(aTools: WchTools, aOut: Writable) {
    // produce the digraph
    aOut.write('digraph g {\n');
    aOut.write(`ranksep = "1.0 equally";\n`);
    // shape config
    aOut.write('node [shape = record];\n');
    // write all sites
    const sites = aTools.sites;
    // tslint:disable-next-line:forin
    for (const siteId in sites) {
        _writeSite(sites[siteId], aTools, aOut);
    }
    // write all content
    const content = aTools.content;
    // tslint:disable-next-line:forin
    for (const contentId in content) {
        _writeContent(content[contentId], aTools, aOut);
    }
    // ok
    aOut.write('}\n');
}

function _showGraph(aOutFile: string, aWchTools: WchTools): Observable<string> {
    // normalize the target
    const outFile = normalize(aOutFile);
    // output stream
    const wrt = createWriteStream(outFile);
    // write
    _writeDigraph(aWchTools, wrt);
    // target
    return of(aOutFile);
}

/**
 * show graph --data <DATA> --out <TARGETNAME>
 */
export class ShowGraphCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'graph';
    static readonly COMMAND = CMD_SHOW;

    constructor(private options: Options) {
        super(options);
    }

    process(): Observable<string> {
        // access to the data folders
        const opts = this.options;
        // out
        const out = resolve(opts.out || './graph.dot');
        // process
        const rxWchTools = this.getWchTools();
        // process
        const rxGraph = rxWchTools.pipe(mergeMap(aTools => _showGraph(out, aTools)));
        // ok
        return rxGraph;
    }
}
