/* Copyright IBM Corp. 2017 */
import { Observable, of, Subject } from 'rxjs';
import { filter, map, mergeMap } from 'rxjs/operators';

import { CMD_SHOW } from './../constants';
import { Options } from './../options';
import { Matcher, WchTools, wchToolsGetMatcher } from './../utils/wchtools';
import { AbstractCommand } from './abstract.command';
import { Command } from './command';

declare type TypeMatcher = (aType: any) => boolean;

function _matches(aName: string, aMatcher: Matcher, aLogger: Subject<string>): boolean {
    // check
    const result = aMatcher(aName);
    aLogger.next(`Matching [${aName}] results in [${result}].`);
    return result;
}

/**
 * Selects components by applying a matching function
 *
 * @param aMatcher  the matcher
 * @param aLogger   the logger
 */
function _byMatcher(aMatcher: Matcher, aLogger: Subject<string>): TypeMatcher {
    // execute the check
    return (aType: any) => aType && aType.name && _matches(aType.name, aMatcher, aLogger);
}

function _listTypes(aTools: WchTools,
    aMatcher: TypeMatcher): Observable<string> {

    // filter the types
    return of(...Object.keys(aTools.types)).pipe(
        map(id => aTools.types[id]),
        filter(aMatcher),
        map(type => type.name)
    );
}

export class ShowTypesCommand extends AbstractCommand implements Command {

    static readonly TYPE = 'types';
    static readonly COMMAND = CMD_SHOW;

    constructor(private options: Options) {
        super(options);
    }

    process(): Observable<string> {
        const opts = this.options;

        // logger
        const logger = this.getLogger();
        // matcher
        const matcher: Matcher = wchToolsGetMatcher(opts);
        const typeMatcher: TypeMatcher = _byMatcher(matcher, logger);

        // tools
        const rxWchTools = this.getWchTools();

        // the types
        const rxTypes = rxWchTools.pipe(mergeMap(aTools => _listTypes(aTools, typeMatcher)));

        // ok
        return rxTypes;
    }
}
